#! /usr/local/bin/perl
# ============================================================================
# $Id: cacti-rrd-import.pl,v 1.0 21/05/2009
# Copyright (c) 2009 Duane Walker
# All rights reserved.
# This program is free software; you may redistribute it and/or modify it
# under the same terms as Perl itself.
# ============================================================================
#
# This script extracts data from the Client Vantage SQL Database.
# It extracts data for the date range specified and invokess the rrdtool
# command line.
#
# perl cacti-rrd-import.pl [-help] | [-start {dd/mm/yyyy|yesterday}] [-end dd/mm/yyyy]
#                          [-app AppName] [-tran TranName] [-debug]
#
# ============================================================================
# This script is designed to extract Application performance stats from
# a Compuware ClientVantage MS SQL Database. It then writes the data to
# RRA files (by invoking RRDTOOL via the command line). The RRA files is
# created if it doesn't exist.
#
# As it steps through the applications, transactions and workstations it
# creates the RRA files, creates the Cacti DataSources, Graphs and Graph
# Trees.
#
# This is specifically written for ClientVantage but can be adapted for any
# data.  Obviously all the SQL queries, naming of files, graphs, data sources
# etc. will need to be changed to something meaningful to you.
#
# It is designed to be run multiple times but always with newer data. It
# reads the last update time from the RRA and won't attempt to put any older
# data int (I don't think RRDTOOL accepts it anyway).
#
# I used ODBC for database access as other versions of this script talk to
# other databases and I wanted to keep the code the same. You can change
# to native Database drivers.
#
# I wrote it on Windows (2000) running Cacti and Perl was provided by
# Cygwin (http://cygwin.com).  I used the cygwin rrdtool binary. I tried
# to get the Perl RRDs or RRDp interfaces working but couldn't build
# RRDTOOL under cygwin. Instead I invoke rrdtool directly.
#
# Initially rrdtool was too slow from the command line (sending one
# value at a time so I construct a command line and then execute a
# batch.  Using this method I could get around 500 values a second.
# RRDTool is so fast!!
#
# Yesterday I built the RRAs and Cacti structure from scratch; it took a
# couple of hours to do over 2 million entries.
#
# Make sure you save you Cacti database.
#
# The main problem I have now is that over time the graph tree items will
# be out of order. I tried a script to re-order them but I haven't been
# able to get it working yet.
#
# Thanks to Cacti, it has an excellent user interface for presenting data.
# I also use it for graphing comms devices.
#
# Thanks to RRDTool, it is so fast and powerful.
#
# ============================================================================
# v1.0 Initial version

use strict;
use DBI;
use Getopt::Long;
use DateTime;

my $timezone = 'Australia/Brisbane';

#-------------------------------------------------------------------------------
#RRDTool defaults
#-------------------------------------------------------------------------------
my $rrdres = 1800;
my $rrdtool = "/usr/share/rrdtool/rrdtool";
my $rrapath = "D:\\Program Files\\Apache Software Foundation\\Apache2.2\\htdocs\\cacti\\rra\\cv";

my $helpFlag = 0;
my $debug = 0;
my $start = '';
my $end = '';
my $app = '';
my $tran = '';
my $localdir = '/cygdrive/d/Perl/';

#-------------------------------------------------------------------------------
#Client Vantage Database details
#-------------------------------------------------------------------------------
my $CVdb = 'prdclientvantage';
my $CVuser = 'xxxxx';
my $CVpass = 'yyyyy';

#-------------------------------------------------------------------------------
#Cacti Database details
#-------------------------------------------------------------------------------
my $cactidb = 'cacti';
my $cactiuser = 'xxxxx';
my $cactipass = 'yyyyy';

#-------------------------------------------------------------------------------
#Variables for external commands
#-------------------------------------------------------------------------------
my $cmd;
my $updateCmd;
my $updateCount;
my $maxUpdateCount = 400;
my @output;
my $exitcode;
my $fail;

GetOptions(
     "help"      	=>		\$helpFlag,
     "start=s"		=> 		\$start,
     "end=s"			=> 		\$end,
     "app=s"			=>		\$app,
     "tran=s"			=>		\$tran,
     "debug"     	=>		\$debug,
	);

usage() if ($helpFlag);

#Initialise start date and end date
my $startDate = DateTime->new( year => 2009, month => 1, day => 1, time_zone => $timezone );
my $endDate = $startDate->clone();
my $now = $startDate->clone();

#-------------------------------------------------------------------------------
#Work out the start and end dates
#-------------------------------------------------------------------------------
if (length($start) == 0){
	#Default to today
	$startDate = DateTime->today();

	#Default end to tomorrow
	#Make endDate of copy of start date and then add 1 day
	$endDate = DateTime->today();
	$endDate->add( hours => 24);
} else {
	#Is the date the string "yesterday"?
	if ($start =~ /^yesterday$/i){
		# yesterday
		$startDate = DateTime->today();

		#End defaults to today
		$endDate = DateTime->today();
	} else {
		if ($start =~ /(\d{1,2})\/(\d{1,2})\/(\d{4})/){
			$startDate->set( year => $3, month => $2, day => $1);

			#End is tomorrow
			$endDate = DateTime->today();
			$endDate->add( hours => 24);
		} elsif ($start =~ /(\d{4})\/(\d{1,2})\/(\d{1,2})/){
			$startDate->set( year => $1, month => $2, day => $3);

			#End defaults to tomorrow
			$endDate = DateTime->today();
			$endDate->add( hours => 24);
		} else {
			print "Incorrect date format, must be dd/mm/yyyy\n";
			usage();
		}
	}
}

if (length($end) > 0){
	if ($end =~ /(\d{1,2})\/(\d{1,2})\/(\d{4})/){
		$endDate->set( year => $3, month => $2, day => $1);

		#We really mean midnight of this date so add 24 hours
		#$endDate->add( hours => 24);
	} elsif ($end =~ /(\d{4})\/(\d{1,2})\/(\d{1,2})/){
		$endDate->set( year => $1, month => $2, day => $3);

		#We really mean midnight of this date so add 24 hours
		#$endDate->add( hours => 24);
	}
}

#Start and end time
my $s = $startDate->ymd('-');
my $e = $endDate->ymd('-');

open LOG, "> cacti-rrd-import.log" or die "Couldn't open Log File cacti-rrd-import.log, $!\n";

$now = DateTime->now;
$now->set_time_zone( $timezone );

print LOG "\n-----------------------------------------------------------------------------------\n";
print LOG "Starting export at ".$now->hms."\n";
print "Starting export at ".$now->hms.' '.$now->epoch."\n";

print 'Start Date: '.$startDate->ymd('/').' End Date: '.$endDate->ymd('/')."\n";
print LOG 'Start Date: '.$startDate->ymd('/').' End Date: '.$endDate->ymd('/')."\n";

#Does the RRA Directory exist?
if (!(-d $rrapath)){
	#No so create it
	($exitcode, @output) = cmd("mkdir \"$rrapath\" ");
	Exit("Couldn't create RRA Directory: \"$rrapath\"\n", @output) if ($exitcode);
}

#-------------------------------------------------------------------------------
#Connect to the Client Vantage Database
#Summarise the transactions in the SQL Database
#-------------------------------------------------------------------------------
my $DSN2 = 'dbi:ODBC:clientvantage';

my $cvDBH = DBI->connect($DSN2, $CVuser, $CVpass);
die "Error connecting to database: $!\n" if (!$cvDBH);

#Delete all the current transactions
$cvDBH->do("use $CVdb;");

#-------------------------------------------------------------------------------
#Connect to the MySQL Server cacti Database via ODBC
#-------------------------------------------------------------------------------
my $DSN3 = 'dbi:ODBC:cacti';
my $cactiDBH = DBI->connect($DSN3, $cactiuser, $cactipass);
die "Error connecting to database: $!\n" if (!$cactiDBH);

#Delete all the current transactions
$cactiDBH->do("use $cactidb;");

#-------------------------------------------------------------------------------
# Build a List of Transaction SLAs and Giveups
#-------------------------------------------------------------------------------
my $query = <<_END_OF_TEXT_;
SELECT A.ApplicationName,T.TransactionName,P.ThresholdTime,P.GiveupTime
FROM PerformanceServiceLevels P
  JOIN Applications A ON P.ApplicationID = A.ApplicationID
    JOIN Transactions T ON P.TransactionID = T.TransactionID
WHERE P.WorkgroupID = 4
ORDER BY A.ApplicationName,T.TransactionName,P.PerformanceSLEffectiveDateUTC
;
_END_OF_TEXT_

my $sth = $cvDBH->prepare($query);
$sth->execute();

my %apps;

my $lastApp = '';
my $lastTran = '';
my $lastAgent = '';

while (my @row = $sth->fetchrow()) {
	my $app = $row[0];
	my $tran = $row[1];

	next if ($tran eq 'RoboticMonitor');
	next if ($tran =~ /^CV_/);
	$tran =~ s/&/-/g;
	$tran =~ s/ //g;

	#print "App $app Tran $tran SLA ".$row[2]." Giveup ".$row[3]."\n" if ($debug);
	$apps{$app}{$tran}{'sla'} = $row[2];
	$apps{$app}{$tran}{'giveup'} = $row[3];
}
$sth->finish();

#-------------------------------------------------------------------------------
# Check whether the heading is defined in Cacti
#-------------------------------------------------------------------------------
#Add a heading for ClientVantage
my $heading = addHeading('ClientVantage');
my $appHeading;
my $appOrderKey;

#-------------------------------------------------------------------------------
# Build a List of Applications and Transactions
#-------------------------------------------------------------------------------
#This query lists the Application, Transaction, Agent, Time and Duration
#-------------------------------------------------------------------------------
$query = <<_END_OF_TEXT_;
SELECT A.ApplicationName,T.TransactionName,AG.AgentName,TD.StartTime,TD.ElapsedTime
FROM TransactionDetails TD
  JOIN Applications A ON TD.ApplicationID = A.ApplicationID
    JOIN Transactions T ON TD.TransactionID = T.TransactionID
      JOIN Agents AG ON TD.AgentID = AG.AgentID
WHERE TD.StartTime BETWEEN '$s 00:00:00' AND '$e 23:59:59'

_END_OF_TEXT_

$query .= " AND A.ApplicationName = '$app' " if (length($app) > 0);
$query .= " AND T.TransactionName = '$tran' " if (length($tran) > 0);
$query .= ' ORDER BY A.ApplicationName,T.TransactionName,AG.AgentName,TD.StartTime;';

my $sth2 = $cvDBH->prepare($query);
$sth2->execute();

my $lastApp = '';
my $lastUpdate;
my $lastTran = '';
my $lastAgent = '';
my $count = 0;


#######################################################################
#Step through the transactions
#Each transaction needs a line in the Update file
#Every combination of App,Tran,Agent will need a rrd file
#######################################################################
while (my @row = $sth2->fetchrow()){
	my $app = $row[0];
	my $tran = $row[1];
	my $agent = $row[2];
	next if ($tran eq 'RoboticMonitor');
	next if ($tran =~ /^CV_/);
	$tran =~ s/&/-/g;
	$tran =~ s/ //g;

	my $rrdName = $tran.'-'.$agent.'.rrd';

	#-----------------------------------------------------------------------------
	#If one of these changes it indicated a new rrd file is required
	#-----------------------------------------------------------------------------
	if (($app ne $lastApp) || ($tran ne $lastTran)  || ($agent ne $lastAgent)){

		#Do we need to add the App to Cacti
		if ($app ne $lastApp){
			#Create a heading for the app (if it already exists it will return the number)
			($appHeading,$appOrderKey) = addAppHeading($app, $heading);
		}

		#Add the Data Source (if it already exists it will return the number)
		#Prefix it with "CV-" and suffix it with the agent name
		my ($ds, $template) = addDataSource($tran."-$agent", 'CV-');
  	
		#print "DataSource $ds Template $template\n" if ($debug);
  	
		#Add a graph (if it already exists it will return the number)
		#Prefix it with "CV-" and suffix it with the agent name
		my 	$gr = addGraph($tran."-$agent", 'CV-',$template, $apps{$app}{$tran}{'sla'}, $apps{$app}{$tran}{'giveup'}, $heading, $appOrderKey);
  	
		#print "Graph $gr\n" if ($debug);

		#Write out the last batch of updates before handling this one
		($exitcode, @output) = cmd($updateCmd) if ($updateCount > 0);

		#Make sure the update command is initialised
		$updateCmd = "$rrdtool update \"$rrapath\\$rrdName\" ";
		$updateCount = 0;

		#Change of App and/or transaction write App/Tran details
		#to create a RRD

		#See if we need to create the RRA file
		if (!(-e $rrapath.'\\'.$rrdName)){
			$cmd = "$rrdtool create \"$rrapath\\$rrdName\" ";
			$cmd .= '--start 1199109600 --step 1800 ';
			$cmd .= 'DS:time:GAUGE:3600:U:U ';
			$cmd .= 'RRA:AVERAGE:0.5:1:672 ';
			$cmd .= 'RRA:AVERAGE:0.5:2:2880 ';
			$cmd .= 'RRA:AVERAGE:0.5:8:2190 ';
			$cmd .= 'RRA:MAX:0.5:1:672 ';
			$cmd .= 'RRA:MAX:0.5:2:2880 ';
			$cmd .= 'RRA:MAX:0.5:8:2190 ';
			($exitcode, @output) = cmd($cmd);
			Exit("Couldn't create RRA File: $rrdName\n", @output) if ($exitcode);
			
			$lastUpdate = 0;
		} else {
			#Find when the last update was
			$cmd = "$rrdtool lastupdate \"$rrapath\\$rrdName\" ";
			($exitcode, @output) = cmd($cmd);
			$lastUpdate = 0;
			if ($exitcode == 0){
				foreach my $line (@output){
					if ($line =~ /^(\d{10}):/){
						$lastUpdate = $1;
						#print "$app $tran $agent - Last Update $lastUpdate\n";
						#print LOG "$app $tran $agent - Last Update $lastUpdate\n";
					}
				}
			}
		}
		print "\n===================================================================\n";
		print "$app $tran $agent\n";
		if (defined($apps{$app}{$tran}{'sla'})){
			print 'SLA '.$apps{$app}{$tran}{'sla'}.' ';
		}
		if (defined($apps{$app}{$tran}{'giveup'})){
			print 'Giveup '.$apps{$app}{$tran}{'giveup'};
		}
		print "\n";
		
		$lastApp = $app;
		$lastTran = $tran;
		$lastAgent = $agent;
	}

	#-----------------------------------------------------------------------------
	#Now write out the transaction details
	#Decode the date and time
	#-----------------------------------------------------------------------------
	$row[3] =~ /(\d{4})-(\d{2})-(\d{2}) (\d{2})\:(\d{2})\:(\d{2})/;
	#print "Year $1 Month $2 Day $3 Hour $4 Minute $5 Second $6\n";
	
	#print $row[2].",".$row[3]."\n";
	my $lineDate = DateTime->new( year => $1, month => $2, day => $3,
																hour => $4, minute => $5, second => $6,
																time_zone => $timezone );

	#-----------------------------------------------------------------------------
	#RRD uses epoch time so convert it
	#Round the time down to the RRD Resolution
	#-----------------------------------------------------------------------------
	my $epoch = $lineDate->epoch();
	$epoch = int($epoch/$rrdres)*$rrdres;		

	#Is this time period before the last update?
	next if ($lastUpdate >= $epoch);
	$lastUpdate = $epoch;

	#Convert the duration from milliseconds to seconds
	my $duration = $row[4]/1000;

	#print $lineDate->dmy.' '.$lineDate->hms.' '.$duration."\n";

	#Batch up the updates to make it faster
	$updateCmd .= " $epoch:$duration";
	$updateCount++;

	#Do we have enough in this batch?
	if ($updateCount >= $maxUpdateCount){
		#Yes so execute the command
		($exitcode, @output) = cmd($updateCmd);

		#Initialise the command for the next batch
		$updateCmd = "$rrdtool update \"$rrapath\\$rrdName\" ";
		$updateCount = 0;
	}
	#Count the transactions and display the count every one thousand
	print $count."\n" if (((++$count) % 1000) == 0);
}

#Write any remaining batch updates
($exitcode, @output) = cmd($updateCmd) if ($updateCount > 0);

$sth2->finish();

#######################################################################
# Now write the transactions out to CSV files
#######################################################################

print "$count rows returned\n";
print LOG "$count rows returned\n";

$cvDBH->disconnect;
$cactiDBH->disconnect;
close LOG;
exit(0);


#######################################################################
# Run an external command and capture the exit code and output
#######################################################################
sub cmd
{
	my ($cmd) = @_;
	
	my @output = `$cmd 2>&1`;
	my $code = $? >> 8;

	if ($code){
		print LOG "--------------------------------------------------------------------------\n";
		print LOG "$cmd\n";
		foreach my $line (@output){
			print $line;
			print LOG $line;
		}
		print LOG "--------------------------------------------------------------------------\n";
	}
	return($code, @output);
}

#######################################################################
# Write the error to the log and exit
#######################################################################
sub Exit
{
	my (@error) = @_;
	foreach (@error){ print LOG; }
	close LOG;
	die $error[0];
}


#######################################################################
#add a Cacti Graph Tree Heading
#######################################################################
sub addHeading
{
	my ($heading) = @_;

	my $query = <<_END_OF_TEXT_;
		SELECT id FROM graph_tree
		WHERE name = '$heading';
_END_OF_TEXT_
	my $sth = $cactiDBH->prepare($query);
	$sth->execute();

	if (my @row = $sth->fetchrow()) {
		#Already exists, nothing to do
		my $num = $row[0];
		$sth->finish();
		return($num);
	}

	#Find the next id (autoincrement)
	my $max_id;
	$sth = $cactiDBH->prepare("select max(id) from graph_tree;");
	$sth->execute;

	if (my @row = $sth->fetchrow()) {
		$max_id = $row[0];
	}
	$sth->finish;


	#Create a new entry in graph_tree
	$sth = $cactiDBH->prepare( q{ INSERT INTO graph_tree
		(id,sort_type,name) 
		VALUES (?,?,?) } );
	$sth->execute($max_id + 1, 4, $heading);
	$sth->finish();

	return($max_id + 1);
}


#######################################################################
#add a Cacti Graph Tree Heading
#######################################################################
sub addAppHeading
{
	my ($app, $heading) = @_;

	my $query = <<_END_OF_TEXT_;
		SELECT id FROM graph_tree_items
		WHERE title = '$app';
_END_OF_TEXT_
	my $sth = $cactiDBH->prepare($query);
	$sth->execute();

	if (my @row = $sth->fetchrow()) {
		#Already exists, nothing to do
		my $num = $row[0];
		$sth->finish();
		return($num);
	}

	my $max_id;
	$sth = $cactiDBH->prepare("select max(id) from graph_tree_items;");
	$sth->execute;

	if (my @row = $sth->fetchrow()) {
		$max_id = $row[0];
	}
	$sth->finish;

	#Find the next menu item number (the orderkey field)
	my $max_okey;
	$sth = $cactiDBH->prepare("select max(order_key) from graph_tree_items where graph_tree_id = $heading;");
	$sth->execute;

	if (my @row = $sth->fetchrow()) {
		$max_okey = substr($row[0], 0, 3);
	}
	$sth->finish;
	my $order_key = sprintf("%03d", $max_okey + 1);

	#Create a new entry in graph_tree
	$sth = $cactiDBH->prepare( q{ INSERT INTO graph_tree_items
		(id,graph_tree_id,local_graph_id,rra_id,title,host_id,order_key,host_grouping_type,sort_children_type) 
		VALUES (?,?,?,?,?,?,?,?,?) } );
	$sth->execute($max_id + 1,$heading,0,0,$app,0,$order_key.'000000000000000000000000000000000000000000000000000000000000000000000000000000000000000',1,2);
	$sth->finish();

	return($max_id + 1, $order_key);
}

#----------------------------------------------------------------------
#Add a DataSource to Cacti
#----------------------------------------------------------------------
sub addDataSource
{
	my ($ds,$prefix) = @_;

	# The DataSource is defined in data_template_data, data_local,
	# data_template_data_rra and data_template_rrd
	# data_templpate_data has the data source name, the data source id is kept in data_local
	# The data source link to the rrd file is in data_template_rrd
	#+-----------------------------+-----------------------+------+-----+---------+----------------+
	#| id                          | mediumint(8) unsigned | NO   | PRI | NULL    | auto_increment |
	#| local_data_template_data_id | mediumint(8) unsigned | NO   |     | 0       |                |
	#| local_data_id               | mediumint(8) unsigned | NO   | MUL | 0       |                |
	#| data_template_id            | mediumint(8) unsigned | NO   | MUL | 0       |                |
	#| data_input_id               | mediumint(8) unsigned | NO   |     | 0       |                |
	#| t_name                      | varchar(2)            | YES  |     | NULL    |                |
	#| name                        | varchar(250)          | NO   |     |         |                |
	#| name_cache                  | varchar(255)          | NO   |     |         |                |
	#| data_source_path            | varchar(255)          | YES  |     | NULL    |                |
	#| t_active                    | varchar(2)            | YES  |     | NULL    |                |
	#| active                      | varchar(2)            | YES  |     | NULL    |                |
	#| t_rrd_step                  | varchar(2)            | YES  |     | NULL    |                |
	#| rrd_step                    | mediumint(8) unsigned | NO   |     | 0       |                |
	#| t_rra_id                    | varchar(2)            | YES  |     | NULL    |                |
	#+-----------------------------+-----------------------+------+-----+---------+----------------+

	#See if the DataSource exists
	my $query = <<_END_OF_TEXT_;
		SELECT dt.local_data_id,dr.id
		FROM data_template_data dt
		  JOIN data_template_rrd dr on dt.local_data_id = dr.local_data_id
		WHERE name = '$prefix$ds';
_END_OF_TEXT_
	my $sth = $cactiDBH->prepare($query);
	$sth->execute();

	if (my @row = $sth->fetchrow()) {
		#Already exists, nothing to do
		my $num1 = $row[0];
		my $num2 = $row[1];
		$sth->finish();
		return($num1, $num2);
	}
	$sth->finish();

	#Need to add a new one
	#Get the current highest id from data_template_data, data_template_rrd
	#and data_local
	my %max_ids;
	my $sth;
	foreach my $table ('data_template_data', 'data_template_rrd', 'data_local'){
		$sth = $cactiDBH->prepare("select max(id) from $table;");
		$sth->execute;

		if (my @row = $sth->fetchrow()) {
			$max_ids{$table} = $row[0];
			#print "$table max id is ".$max_ids{$table}."\n";
		}

		$sth->finish;
	}

	#Assign the new record ids
	my $local_data_id = $max_ids{'data_local'} + 1;
	my $data_template_data_id = $max_ids{'data_template_data'} + 1;
	my $data_template_rrd_id = $max_ids{'data_template_rrd'} + 1;

	#-----------------------------------------------------------------------------
	#Create a new entry in data_local
	#This table doesn't do much for our purpose (the settings are all blank)
	#but is needed so the datasource displays correctly in Cacti
	#+------------------+-----------------------+------+-----+---------+----------------+
	#| id               | mediumint(8) unsigned | NO   | PRI | NULL    | auto_increment |
	#| data_template_id | mediumint(8) unsigned | NO   |     | 0       |                |
	#| host_id          | mediumint(8) unsigned | NO   | MUL | 0       |                |
	#| snmp_query_id    | mediumint(8)          | NO   |     | 0       |                |
	#| snmp_index       | varchar(255)          | NO   |     |         |                |
	#+------------------+-----------------------+------+-----+---------+----------------+
	$sth = $cactiDBH->prepare( q{ INSERT INTO data_local 
			(id,data_template_id,host_id,snmp_query_id,snmp_index) 
			VALUES (?,?,?,?,?) } );
	$sth->execute($local_data_id, 0, 0, 0, '' );
	die "Error inserting 'data_local_id'- ".$local_data_id." ".$sth->errstr()."\n" if ($sth->err);
	$sth->finish();
	
	#-----------------------------------------------------------------------------
	#Create a new entry in data_template_data
	#This contains the name and some of the data source details
	#+-----------------------------+-----------------------+------+-----+---------+----------------+
	#| id                          | mediumint(8) unsigned | NO   | PRI | NULL    | auto_increment |
	#| local_data_template_data_id | mediumint(8) unsigned | NO   |     | 0       |                |
	#| local_data_id               | mediumint(8) unsigned | NO   | MUL | 0       |                |
	#| data_template_id            | mediumint(8) unsigned | NO   | MUL | 0       |                |
	#| data_input_id               | mediumint(8) unsigned | NO   |     | 0       |                |
	#| t_name                      | varchar(2)            | YES  |     | NULL    |                |
	#| name                        | varchar(250)          | NO   |     |         |                |
	#| name_cache                  | varchar(255)          | NO   |     |         |                |
	#| data_source_path            | varchar(255)          | YES  |     | NULL    |                |
	#| t_active                    | varchar(2)            | YES  |     | NULL    |                |
	#| active                      | varchar(2)            | YES  |     | NULL    |                |
	#| t_rrd_step                  | varchar(2)            | YES  |     | NULL    |                |
	#| rrd_step                    | mediumint(8) unsigned | NO   |     | 0       |                |
	#| t_rra_id                    | varchar(2)            | YES  |     | NULL    |                |
	#+-----------------------------+-----------------------+------+-----+---------+----------------+
	$sth = $cactiDBH->prepare( q{ INSERT INTO data_template_data
		 (id,local_data_template_data_id,local_data_id,data_template_id,data_input_id,t_name,name,name_cache,data_source_path,t_active,active,t_rrd_step,rrd_step,t_rra_id) 
		 VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?) } );
	$sth->execute($data_template_data_id, 0, $local_data_id, 0, 0, undef, $prefix.$ds, $prefix.$ds, "<path_rra>/cv/$ds\.rrd", undef, '', undef, 1800, undef);
	die "Error inserting 'data_template_data_id'- ".$data_template_data_id." ".$sth->errstr()."\n" if ($sth->err);
	$sth->finish();
	
	#-----------------------------------------------------------------------------
	#Create a new entry in data_template_data_rra
	#This would contain details of how the rrds would be created but 
	#Cacti doesn't create them anyway. For consistency the standard
	#Cacti records are added.
	#+-----------------------+-----------------------+------+-----+---------+-------+
	#| Field                 | Type                  | Null | Key | Default | Extra |
	#+-----------------------+-----------------------+------+-----+---------+-------+
	#| data_template_data_id | mediumint(8) unsigned | NO   | PRI | 0       |       |
	#| rra_id                | mediumint(8) unsigned | NO   | PRI | 0       |       |
	#+-----------------------+-----------------------+------+-----+---------+-------+
	$sth = $cactiDBH->prepare( q{ INSERT INTO data_template_data_rra
		(data_template_data_id,rra_id) 
		VALUES (?,?) } );
	for (my $i = 1; $i <= 5; $i++){
		$sth->execute($data_template_data_id, $i);
		die "Error inserting 'data_template_data_id'- ".$data_template_data_id.." ".$i." ".$sth->errstr()."\n" if ($sth->err);
	}
	$sth->finish();
	
	#-----------------------------------------------------------------------------
	#Create a new entry in data_template_rrd
	#This links the datasource to the rrd file and has some rra settings
	#+----------------------------+-----------------------+------+-----+---------+----------------+
	#| id                         | mediumint(8) unsigned | NO   | PRI | NULL    | auto_increment |
	#| hash                       | varchar(32)           | NO   |     |         |                |
	#| local_data_template_rrd_id | mediumint(8) unsigned | NO   |     | 0       |                |
	#| local_data_id              | mediumint(8) unsigned | NO   | MUL | 0       |                |
	#| data_template_id           | mediumint(8) unsigned | NO   | MUL | 0       |                |
	#| t_rrd_maximum              | varchar(2)            | YES  |     | NULL    |                |
	#| rrd_maximum                | varchar(20)           | NO   |     | 0       |                |
	#| t_rrd_minimum              | varchar(2)            | YES  |     | NULL    |                |
	#| rrd_minimum                | varchar(20)           | NO   |     | 0       |                |
	#| t_rrd_heartbeat            | varchar(2)            | YES  |     | NULL    |                |
	#| rrd_heartbeat              | mediumint(6)          | NO   |     | 0       |                |
	#| t_data_source_type_id      | varchar(2)            | YES  |     | NULL    |                |
	#| data_source_type_id        | smallint(5)           | NO   |     | 0       |                |
	#| t_data_source_name         | varchar(2)            | YES  |     | NULL    |                |
	#| data_source_name           | varchar(19)           | NO   |     |         |                |
	#| t_data_input_field_id      | varchar(2)            | YES  |     | NULL    |                |
	#| data_input_field_id        | mediumint(8) unsigned | NO   |     | 0       |                |
	#+----------------------------+-----------------------+------+-----+---------+----------------+

	$sth = $cactiDBH->prepare( q{ INSERT INTO data_template_rrd
		 (id,hash,local_data_template_rrd_id,local_data_id,data_template_id,t_rrd_maximum,rrd_maximum,t_rrd_minimum,rrd_minimum,t_rrd_heartbeat,rrd_heartbeat,t_data_source_type_id,data_source_type_id,t_data_source_name,data_source_name,t_data_input_field_id,data_input_field_id) 
		 VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?) } );
	$sth->execute($data_template_rrd_id,'',0,$local_data_id,0,undef,'0',undef,'0',undef,3600,undef,1,undef,'time',undef,0);
	die "Error inserting 'data_template_rrd_id'- ".$data_template_rrd_id." ".$sth->errstr()."\n" if ($sth->err);
	$sth->finish();

	return($local_data_id, $data_template_rrd_id);
}


#----------------------------------------------------------------------
#Add a Cacti Graph
#----------------------------------------------------------------------
sub addGraph
{
	my ($gr, $prefix, $ds, $sla, $giveup, $heading, $appOrderKey) = @_;

	#-----------------------------------------------------------------------------
	#The graphs are defined in graph_templates_graph, graph_local, graph_templates_item
	#-----------------------------------------------------------------------------
	#See if this graph is already defined
	my $query = <<_END_OF_TEXT_;
		SELECT gt.id FROM graph_templates_graph gt
		WHERE title = '$prefix$gr';
_END_OF_TEXT_
	my $sth = $cactiDBH->prepare($query);
	$sth->execute();

	if (my @row = $sth->fetchrow()) {
		#Already exists, nothing to do
		my $num = $row[0];
		$sth->finish();
		return($num);
	}
	$sth->finish();

	#Need to add the graph
	#Get the current highest id from graph_local, graph_templates_graph,
	#graph_templates_item, graph_tree_items
	my %max_ids;
	my $sth;
	foreach my $table ('graph_local', 'graph_templates_graph', 'graph_templates_item', 'graph_tree_items'){
		$sth = $cactiDBH->prepare("select max(id) from $table;");
		$sth->execute;

		if (my @row = $sth->fetchrow()) {
			$max_ids{$table} = $row[0];
			#print "$table max id is ".$max_ids{$table}."\n";
		}
		$sth->finish;
	}

	#Assign the new record ids
	my $local_graph_id = $max_ids{'graph_local'} + 1;
	my $graph_templates_graph_id = $max_ids{'graph_templates_graph'} + 1;
	my $graph_templates_item_id = $max_ids{'graph_templates_item'} + 1;
	my $graph_tree_items_id = $max_ids{'graph_tree_items'} + 1;

	#-----------------------------------------------------------------------------	
	#Create a new entry in graph_local
	#graph_local, for our pusposes, only has the graph id in it, the rest of
	#the settings are blank.
	#+-------------------+-----------------------+------+-----+---------+----------------+
	#| id                | mediumint(8) unsigned | NO   | PRI | NULL    | auto_increment |
	#| graph_template_id | mediumint(8) unsigned | NO   | MUL | 0       |                |
	#| host_id           | mediumint(8) unsigned | NO   | MUL | 0       |                |
	#| snmp_query_id     | mediumint(8)          | NO   |     | 0       |                |
	#| snmp_index        | varchar(255)          | NO   | MUL |         |                |
	#+-------------------+-----------------------+------+-----+---------+----------------+
	$sth = $cactiDBH->prepare( q{ INSERT INTO graph_local 
			(id,graph_template_id,host_id,snmp_query_id,snmp_index) 
			VALUES (?,?,?,?,?) } );
	$sth->execute($local_graph_id, 0, 0, 0, '' );
	die "Error inserting 'local_graph_id'- ".$local_graph_id." ".$sth->errstr()."\n" if ($sth->err);
	$sth->finish();

	#-----------------------------------------------------------------------------	
	#Create a new entry in graph_templates_graph
	#graph_templates_graph has the overall graph settings with each of the items
	#(lines, text, etc) in graph_templates_item
	#+-------------------------------+-----------------------+------+-----+---------+----------------+
	#| id                            | mediumint(8) unsigned | NO   | PRI | NULL    | auto_increment |
	#| local_graph_template_graph_id | mediumint(8) unsigned | NO   |     | 0       |                |
	#| local_graph_id                | mediumint(8) unsigned | NO   | MUL | 0       |                |
	#| graph_template_id             | mediumint(8) unsigned | NO   | MUL | 0       |                |
	#| t_image_format_id             | varchar(2)            | YES  |     | 0       |                |
	#| image_format_id               | tinyint(1)            | NO   |     | 0       |                |
	#| t_title                       | varchar(2)            | YES  |     | 0       |                |
	#| title                         | varchar(255)          | NO   |     |         |                |
	#| title_cache                   | varchar(255)          | NO   | MUL |         |                |
	#| t_height                      | varchar(2)            | YES  |     | 0       |                |
	#| height                        | mediumint(8)          | NO   |     | 0       |                |
	#| t_width                       | varchar(2)            | YES  |     | 0       |                |
	#| width                         | mediumint(8)          | NO   |     | 0       |                |
	#| t_upper_limit                 | varchar(2)            | YES  |     | 0       |                |
	#| upper_limit                   | varchar(20)           | NO   |     | 0       |                |
	#| t_lower_limit                 | varchar(2)            | YES  |     | 0       |                |
	#| lower_limit                   | varchar(20)           | NO   |     | 0       |                |
	#| t_vertical_label              | varchar(2)            | YES  |     | 0       |                |
	#| vertical_label                | varchar(200)          | YES  |     | NULL    |                |
	#| t_slope_mode                  | varchar(2)            | YES  |     | 0       |                |
	#| slope_mode                    | varchar(2)            | YES  |     | on      |                |
	#| t_auto_scale                  | varchar(2)            | YES  |     | 0       |                |
	#| auto_scale                    | varchar(2)            | YES  |     | NULL    |                |
	#| t_auto_scale_opts             | varchar(2)            | YES  |     | 0       |                |
	#| auto_scale_opts               | tinyint(1)            | NO   |     | 0       |                |
	#| t_auto_scale_log              | varchar(2)            | YES  |     | 0       |                |
	#| auto_scale_log                | varchar(2)            | YES  |     | NULL    |                |
	#| t_scale_log_units             | varchar(2)            | YES  |     | 0       |                |
	#| scale_log_units               | char(2)               | YES  |     | NULL    |                |
	#| t_auto_scale_rigid            | varchar(2)            | YES  |     | 0       |                |
	#| auto_scale_rigid              | varchar(2)            | YES  |     | NULL    |                |
	#| t_auto_padding                | varchar(2)            | YES  |     | 0       |                |
	#| auto_padding                  | varchar(2)            | YES  |     | NULL    |                |
	#| t_base_value                  | varchar(2)            | YES  |     | 0       |                |
	#| base_value                    | mediumint(8)          | NO   |     | 0       |                |
	#| t_grouping                    | varchar(2)            | YES  |     | 0       |                |
	#| grouping                      | varchar(2)            | NO   |     |         |                |
	#| t_export                      | varchar(2)            | YES  |     | 0       |                |
	#| export                        | varchar(2)            | YES  |     | NULL    |                |
	#| t_unit_value                  | varchar(2)            | YES  |     | 0       |                |
	#| unit_value                    | varchar(20)           | YES  |     | NULL    |                |
	#| t_unit_exponent_value         | varchar(2)            | YES  |     | 0       |                |
	#| unit_exponent_value           | varchar(5)            | NO   |     |         |                |
	#+-------------------------------+-----------------------+------+-----+---------+----------------+
	$sth = $cactiDBH->prepare( q{ INSERT INTO graph_templates_graph 
			(id,local_graph_template_graph_id,local_graph_id,graph_template_id,
				t_image_format_id,image_format_id,t_title,title,title_cache,
				t_height,height,t_width,width,t_upper_limit,upper_limit,t_lower_limit,
				lower_limit,t_vertical_label,vertical_label,t_slope_mode,slope_mode,
				t_auto_scale,auto_scale,t_auto_scale_opts,auto_scale_opts,t_auto_scale_log,
				auto_scale_log,t_scale_log_units,scale_log_units,t_auto_scale_rigid,
				auto_scale_rigid,t_auto_padding,auto_padding,t_base_value,base_value,
				t_grouping,grouping,t_export,export,t_unit_value,unit_value,t_unit_exponent_value,
				unit_exponent_value) 
			VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?) } );
	$sth->execute($graph_templates_graph_id,0,$local_graph_id,0,'0',1,'0',$prefix.$gr,$prefix.$gr,'0',120,'0',500,'0','100','0','0','0','seconds','0','on','0','on','0',2,'0','','0','','0','','0','on','0',1000,'0','','0','on','0','','0','');
	die "Error inserting 'local_graph_id'- ".$local_graph_id." ".$sth->errstr()."\n" if ($sth->err);
	$sth->finish();

	#-----------------------------------------------------------------------------	
	#Now add Graph Items (area line or stacked plot), calculations, hlines etc
	#+------------------------------+-----------------------+------+-----+---------+----------------+
	#| id                           | int(12) unsigned      | NO   | PRI | NULL    | auto_increment |
	#| hash                         | varchar(32)           | NO   |     |         |                |
	#| local_graph_template_item_id | int(12) unsigned      | NO   |     | 0       |                |
	#| local_graph_id               | mediumint(8) unsigned | NO   | MUL | 0       |                |
	#| graph_template_id            | mediumint(8) unsigned | NO   | MUL | 0       |                |
	#| task_item_id                 | mediumint(8) unsigned | NO   |     | 0       |                |
	#| color_id                     | mediumint(8) unsigned | NO   |     | 0       |                |
	#| alpha                        | varchar(2)            | YES  |     | FF      |                |
	#| graph_type_id                | tinyint(3)            | NO   |     | 0       |                |
	#| cdef_id                      | mediumint(8) unsigned | NO   |     | 0       |                |
	#| consolidation_function_id    | tinyint(2)            | NO   |     | 0       |                |
	#| text_format                  | varchar(255)          | YES  |     | NULL    |                |
	#| value                        | varchar(255)          | YES  |     | NULL    |                |
	#| hard_return                  | varchar(2)            | YES  |     | NULL    |                |
	#| gprint_id                    | mediumint(8) unsigned | NO   |     | 0       |                |
	#| sequence                     | mediumint(8) unsigned | NO   |     | 0       |                |
	#+------------------------------+-----------------------+------+-----+---------+----------------+
	#Firstly add the area graph of the transaction times
	$sth = $cactiDBH->prepare( q{ INSERT INTO graph_templates_item 
			(id,hash,local_graph_template_item_id,local_graph_id,graph_template_id,
				task_item_id,color_id,alpha,graph_type_id,cdef_id,
				consolidation_function_id,text_format,value,hard_return,gprint_id,sequence
			) 
			VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?) } );
	#(411,'',0,5,0,97,17,'FF',7,0,1,'','','',2,1)
	$sth->execute($graph_templates_item_id,'',0,$local_graph_id,0,$ds,17,'FF',7,0,1,'','','',2,1);
	$graph_templates_item_id++;
	die "Error inserting 'graph_templates_item_id'- ".$graph_templates_item_id." ".$sth->errstr()."\n" if ($sth->err);

	#Now add a HLine for the SLA time
	#SLA
	#(412,'',0,5,0,97,31,'FF',2,0,1,'SLA','8','',2,2)
	if ($sla > 0){
		$sth->execute($graph_templates_item_id,'',0,$local_graph_id,0,$graph_templates_graph_id,31,'FF',2,0,1,'SLA',$sla,'',2,2);
		$graph_templates_item_id++;

		#Add a comment item as well
		#+------+------+------------------------------+----------------+-------------------+--------------+----------+-------+---------------+---------+---------------------------+-------------------+-------+-------------+-----------+----------+
		#| id   | hash | local_graph_template_item_id | local_graph_id | graph_template_id | task_item_id | color_id | alpha | graph_type_id | cdef_id | consolidation_function_id | text_format       | value | hard_return | gprint_id | sequence |
		#+------+------+------------------------------+----------------+-------------------+--------------+----------+-------+---------------+---------+---------------------------+-------------------+-------+-------------+-----------+----------+
		#| 3502 |      |                            0 |           1406 |                 0 |         1453 |        9 | FF    |             2 |       0 |                         1 | GiveUp            | 30    |             |         2 |        3 |
		#| 3501 |      |                            0 |           1406 |                 0 |         1453 |       31 | FF    |             2 |       0 |                         1 | SLA               | 5     |             |         2 |        2 |
		#| 3500 |      |                            0 |           1406 |                 0 |         1502 |       17 | FF    |             7 |       0 |                         1 |                   |       |             |         2 |        1 |
		#| 3607 |      |                            0 |           1406 |                 0 |            0 |        0 | FF    |             1 |       0 |                         1 | SLA 5 Seconds     |       |             |         2 |        4 |
		#| 3608 |      |                            0 |           1406 |                 0 |            0 |        0 | FF    |             1 |       0 |                         1 | Giveup 30 Seconds |       |             |         2 |        5 |
		#+------+------+------------------------------+----------------+-------------------+--------------+----------+-------+---------------+---------+---------------------------+-------------------+-------+-------------+-----------+----------+
		$sth->execute($graph_templates_item_id,'',0,$local_graph_id,0,0,0,'FF',1,0,1,'SLA '.$sla.' seconds','','',2,3);
		$graph_templates_item_id++;
	}

	#Now add a HLine for the Giveup time
	#Giveup
	#(413,'',0,5,0,97,9,'FF',2,0,1,'GiveUp','30','',2,3)
	if ($giveup > 0){
		$sth->execute($graph_templates_item_id,'',0,$local_graph_id,0,$graph_templates_graph_id,9,'FF',2,0,1,'GiveUp',$giveup,'',2,4);
		$graph_templates_item_id++;

		#Add a comment item as well
		#+------+------+------------------------------+----------------+-------------------+--------------+----------+-------+---------------+---------+---------------------------+-------------------+-------+-------------+-----------+----------+
		#| id   | hash | local_graph_template_item_id | local_graph_id | graph_template_id | task_item_id | color_id | alpha | graph_type_id | cdef_id | consolidation_function_id | text_format       | value | hard_return | gprint_id | sequence |
		#+------+------+------------------------------+----------------+-------------------+--------------+----------+-------+---------------+---------+---------------------------+-------------------+-------+-------------+-----------+----------+
		#| 3502 |      |                            0 |           1406 |                 0 |         1453 |        9 | FF    |             2 |       0 |                         1 | GiveUp            | 30    |             |         2 |        3 |
		#| 3501 |      |                            0 |           1406 |                 0 |         1453 |       31 | FF    |             2 |       0 |                         1 | SLA               | 5     |             |         2 |        2 |
		#| 3500 |      |                            0 |           1406 |                 0 |         1502 |       17 | FF    |             7 |       0 |                         1 |                   |       |             |         2 |        1 |
		#| 3607 |      |                            0 |           1406 |                 0 |            0 |        0 | FF    |             1 |       0 |                         1 | SLA 5 Seconds     |       |             |         2 |        4 |
		#| 3608 |      |                            0 |           1406 |                 0 |            0 |        0 | FF    |             1 |       0 |                         1 | Giveup 30 Seconds |       |             |         2 |        5 |
		#+------+------+------------------------------+----------------+-------------------+--------------+----------+-------+---------------+---------+---------------------------+-------------------+-------+-------------+-----------+----------+
		$sth->execute($graph_templates_item_id,'',0,$local_graph_id,0,0,0,'FF',1,0,1,'Giveup '.$giveup.' seconds','','',2,5);
		$graph_templates_item_id++;
	}
	$sth->finish();

	#-----------------------------------------------------------------------------	
	#Now add the graph to the graph tree App Heading
	#+--------------------+-----------------------+------+-----+---------+----------------+
	#| id                 | mediumint(5) unsigned | NO   | PRI | NULL    | auto_increment |
	#| graph_tree_id      | smallint(5) unsigned  | NO   | MUL | 0       |                |
	#| local_graph_id     | mediumint(8) unsigned | NO   | MUL | 0       |                |
	#| rra_id             | smallint(8) unsigned  | NO   |     | 0       |                |
	#| title              | varchar(255)          | YES  |     | NULL    |                |
	#| host_id            | mediumint(8) unsigned | NO   | MUL | 0       |                |
	#| order_key          | varchar(100)          | NO   | MUL | 0       |                |
	#| host_grouping_type | tinyint(3) unsigned   | NO   |     | 1       |                |
	#| sort_children_type | tinyint(3) unsigned   | NO   |     | 1       |                |
	#+--------------------+-----------------------+------+-----+---------+----------------+
	#The order key specifies the order of menu items within a heading
	#It is thirty digits long with each set of three digits giving the position withina  level.
	#It therefore allows for 10 levels of headings (plus one extra level for the main heading).
	#To find where to add it look at the relevant level (2) and find the maximum value
	#within that level.
	$sth = $cactiDBH->prepare( q{ Select max(order_key) from graph_tree_items where graph_tree_id = '$appOrderKey'; } );
	$sth->execute();

	$max_ids{'order_key'} = 1;
	if (my @row = $sth->fetchrow()) {
		#Already exists, nothing to do
		$max_ids{'order_key'} = substr($row[0],3,3);
	}
	$sth->finish();
	#The new order key is incremented and formatted as a string
	my $ok = sprintf("%03d", $max_ids{'order_key'} + 1);

	#-----------------------------------------------------------------------------	
	#Now create the heading and add the graph to it
	#-----------------------------------------------------------------------------	
	$sth = $cactiDBH->prepare( q{ INSERT INTO graph_tree_items 
			(id,graph_tree_id,local_graph_id,rra_id,title,host_id,order_key,host_grouping_type,sort_children_type) 
			VALUES (?,?,?,?,?,?,?,?,?) } );
	$sth->execute($graph_tree_items_id,$heading,$local_graph_id,5,'',0,$appOrderKey.$ok.'000000000000000000000000000000000000000000000000000000000000000000000000000000000000',1,1);
	die "Error inserting 'graph_tree_items_id'- ".$graph_tree_items_id." ".$sth->errstr()."\n" if ($sth->err);
	$sth->finish();
	
	return($local_graph_id);
}


#######################################################################
# Command line options processing
#######################################################################
sub usage(){

my $usage = <<_END_OF_TEXT_;

This script extracts data from the Client Vantage SQL Database.
It extracts data for the date range specified and invokes RRDTool
to update RRD files on the cacti server.

It extracts for the date range specified.

usage: $0 [-help] | {-start dd/mm/yyyy} {-end dd/mm/yyyy}
                  [-app appName [-tran transactionName]]
                  [debug]

 -help          : display this text
 -start         : start date dd/mm/yyyy (defaults to today)
 -end           : end date dd/mm/yyyy (defaults to tomorrow)
 -app           : application name (optional)
 -tran          : transaction name (optional)
 -debug         : display debugging info

examples:

$0 -help
$0 -start 01/05/2009 
$0 -start 01/05/2009 -end 07/05/2009 -app WebApp
$0 -start 01/05/2009 -end 07/05/2009 -app WebApp -tran WebApp_Homepage
$0 -start 01/05/2009 -debug

The command line options can be shortened to unique values.
eg. -user can be shortened to -u

_END_OF_TEXT_

    print STDERR $usage;
    exit;
}
